source("estimate_theta.R")

select_item <- function(theta, items_used, item_params, n_dims, n_cats, D, Q, current_len, max_len, count, lower, upper) {
  
  ##############################################################################
  # select the next item with constraint-weighted index
  ##############################################################################
  # Paramters
  # theta: simulee's current theta estimates
  # item_used: a vector of item no.
  # item_params: a matrix of item parameters, including the used items
  # n_dims: no. of latent trait dimensions
  # n_cats: a vector that describes the number of categories for each item
  # D: scaling constant, 1 for logistic model, 1.7 for normal ogive model
  # Q: a matrix that indicates which item is related to which dimension
  # current_len: current test length
  # max_len: maximum test length
  # lower: minimum number of items from each latent trait dimension
  # upper: maximum number of items from each latent trait dimension
  #
  # Returns
  # next_item: the index of the item to be administered next
  ##############################################################################
  
  n_items <- nrow(item_params)
  a <- item_params[, 1:n_dims]
  b <- item_params[, (n_dims + 1):ncol(item_params)]
  items_used <- items_used[1:(current_len - 1)]
  items_unused <- setdiff(1:n_items, items_used)
  
  # compute the Fisher information matrix for all items
  fi_mat <- array(0, c(n_dims, n_dims, n_items))
  for (i in 1:n_items) {
    g <- n_cats[i]
    cum_P <- sapply(1:(g - 1), function(h) 1 / (1 + exp(-D * a[i, ] %*% (theta - b[i, h]))))
    cum_P <- c(1, cum_P, 0)
    P <- cum_P[1:g] - cum_P[2:(g + 1)]
    temp <- sum((cum_P[1:g] * (1 - cum_P[1:g]) - cum_P[2:(g + 1)] * (1 - cum_P[2:(g + 1)])) ^ 2 / P)
    fi_mat[, , i] <- D ^ 2 * tcrossprod(a[i, ], a[i, ]) * temp
  }
  
  if (current_len <= n_dims) {
    # for the first few items, manually select an item from each dimension
    # as the determinant of the accumulated Fisher information matrix is zero
    fi_1d <- cbind(1:nrow(item_params), apply(fi_mat, 3, sum))
    fi_1d <- fi_1d[Q[, current_len] == 1, ]
    next_item <- fi_1d[order(fi_1d[, 2], decreasing = TRUE)[1], 1]
  } else {
    # compute the sum of Fisher information matrices of items that are already administered
    fi_used_items <- array(0, c(n_dims, n_dims))
    for (i in items_used)
      fi_used_items <- fi_used_items + fi_mat[, , i]
    
    # compute the determinants of the sum of Fisher information matrices of 
    # already administered items plus each of the unused items
    d <- sapply(items_unused, function(i) det(fi_used_items + fi_mat[, , i]))
    
    # compute the constraint-weighted index
    f1 <- (upper - count - 1) / upper
    f2 <- (max_len - lower - current_len + count) / (max_len - lower)
    weight <- f1 * f2
    
    # commpute the selection criterion
    criterion <- d * tcrossprod(weight, Q[items_unused, ])
    
    max_criterion <- which(criterion == max(criterion))
    if (length(max_criterion) == 1) next_item <- items_unused[max_criterion]
    else next_item <- items_unused[sample(max_criterion, 1)]
  }
  return(next_item)
}


cat_sim <- function(true_theta, initial_theta, item_params, n_dims, n_cats, D, max_len, lower, upper) {
  
  ##############################################################################
  # computerized adaptive testing for Multidimensional Graded Response Model
  ##############################################################################
  # Paramters
  # true_theta: simulee's true theta level
  # initial_theta: initial theta level used for selecting the first item
  # item_params: a matrix of item parametrs (data.frame doesn't work);
  #              first k cols are assumed to be a parameters where k is no. of dimensions;
  #              the other cols are assumed to be b parameters
  # n_dims: no. of latent trait dimensions
  # n_cats: a vector that describes the number of categories for each item
  # D: scaling constant, 1 for logistic model, 1.7 for normal ogive model
  # max_len: maximum test length
  # lower: a vector containing the minimum no. of items to be administered from each latent trait dimension
  # upper: a vector containing the maximum no. of items to be administered from each latent trait dimension
  #
  # Returns
  # a matrix that includes
  # item_no: the order of the items being administered
  # theta_hat: theta estimates for each latent trait dimension
  # std_error: standard errors for each latent trait dimension
  # items_used: indices of the items used in the test
  # resp: responses for used item, note that the response to the first category is 0, not 1
  # det_fi: the determinant of the observed fisher information matrix
  ##############################################################################
  
  theta_hat <- std_error <- array(dim = c(max_len, n_dims))
  items_used <- resp <- det_fi <- array(dim = max_len)
  Q <- sign(abs(item_params[, 1:n_dims])) # a matrix that indicates which item is related to which dimension
  count <- rep(0, n_dims) # count how many items have been administered for each latent trait dimension
  
  for (i in 1:max_len) {
    if (i == 1) theta <- initial_theta
    else theta <- theta_hat[i - 1, ]
    
    # select the next item based on the current theta estimates
    next_item <- select_item(theta, items_used, item_params, n_dims, n_cats, D, Q, i, max_len, count, lower, upper)
    items_used[i] <- next_item # record which item is selected
    count <- count + Q[next_item, ] # update the no. of items that has been administered for each dimension
    
    # simulate a response for the selected item based on true theta
    a <- item_params[next_item, 1:n_dims]
    b <- item_params[next_item, (n_dims + 1):ncol(item_params)]
    cum_P <- sapply(1:(n_cats[next_item] - 1), function(h) 1 / (1 + exp(-D * a %*% (true_theta - b[h]))))
    resp[i] <- sum(cum_P > runif(1))
    
    # compute theta estimates and standard errors
    if (i < n_dims) { # can't estimate theta until we have responses for each dimension
      theta_hat[i, ] <- rep(NA, n_dims)
      std_error[i, ] <- rep(NA, n_dims)
      det_fi[i] <- 0
    } else {
      est <- estimate_theta(item_params[items_used[1:i], , drop = F], resp[1:i], n_dims, D)
      theta_hat[i, ] <- est$theta_hat
      std_error[i, ] <- est$std_error
      det_fi[i] <- est$det_fi
    }
  }
  
  # output all relevant information into a matrix
  item_no <- 1:i
  colnames(theta_hat) <- paste0("theta_hat", 1:n_dims)
  colnames(std_error) <- paste0("std_error", 1:n_dims)
  out <- cbind(item_no, round(theta_hat, 7), round(std_error, 7), items_used, resp, det_fi)
  return(out)
}
