##############################################################################
# omnibus test (LRT, MWT, MMWT and ST)
# LRT: likelihood ratio test
# MWT: multivariate Wald test
# MMWT: modified multivariate Wald test
# ST: score test
##############################################################################
# Parameters
# item_params: a 3-dimensional array (1: items, 2: parameters, 3: occasions)
# resp: a 3-dimensional array (1: people, 2: items, 3: occasions)
# theta_pooled: a 2-dimensional array (1: people, 2: dimension)
# theta_occasion: a 3-dimensional array (1: people, 2: dimension, 3: occasions)
# n_dims: no. of latent trait dimensions
# n_occasions: no. of occasions
# D: scaling constant
# alpha: default is 0.05
#
# Returns
# a list containing the test statistic, the p-value and the deicision
##############################################################################

source("log_lik.R")

LRT_omnibus <- function(item_params, resp, theta_pooled, theta_occasion, n_dims, n_occasions, D, alpha = 0.05) {
  log_lik_null <- log_lik_alt <- 0
  for (t in 1:n_occasions) {
    log_lik_null <- log_lik_null + log_lik(item_params[, , t], resp[, t], n_dims, D, theta_pooled)
    log_lik_alt <- log_lik_alt + log_lik(item_params[, , t], resp[, t], n_dims, D, theta_occasion[t, ])
  }
  
  test_statistic <- -2 * (log_lik_null - log_lik_alt)
  df <- n_dims * (n_occasions - 1)
  p_value <- pchisq(test_statistic, df, lower.tail = FALSE)
  decision <- ifelse(p_value < alpha, 1, 0)
  return(list(test_statistic = test_statistic, p_value = p_value, decision = decision))
}


MWT_omnibus <- function(item_params, resp, theta_pooled, theta_occasion, n_dims, n_occasions, D, alpha = 0.05) {
  # compute observed information matrices
  obs_fi <- array(0, c(n_dims, n_dims, n_occasions))
  for (t in 1:n_occasions)
    obs_fi[, , t] <- -log_lik(item_params[, , t], resp[, t], n_dims, D, theta_pooled, order = 2)
  
  test_statistic <- 0
  for (t1 in 1:n_occasions) {
    for (t2 in 1:n_occasions) {
      if (t2 > t1) {
        diff <- theta_occasion[t1, ] - theta_occasion[t2, ]
        var_diff <- solve(solve(obs_fi[, , t1]) + solve(obs_fi[, , t2]))
        test_statistic <- test_statistic + diff %*% var_diff %*% diff
      }
    }
  }
  
  df <- n_dims * n_occasions * (n_occasions - 1) / 2
  p_value <- pchisq(test_statistic, df, lower.tail = FALSE)
  decision <- ifelse(p_value < alpha, 1, 0)
  return(list(test_statistic = test_statistic, p_value = p_value, decision = decision))
}


MMWT_omnibus <- function(item_params, resp, theta_pooled, theta_occasion, n_dims, n_occasions, D, alpha = 0.05) {
  if (n_occasions != 4) stop("Number of occasions has to be 4")
  
  # compute observed information matrices
  obs_fi <- array(0, c(n_dims, n_dims, n_occasions))
  for (t in 1:n_occasions)
    obs_fi[, , t] <- -log_lik(item_params[, , t], resp[, t], n_dims, D, theta_pooled, order = 2)
  
  linear_coef <- rbind(c(1/3, 1/3, 1/3, -1.0),
                       c(1/2, 1/2, -1.0, 0.0),
                       c(1.0, -1.0, 0.0, 0.0))
  
  test_statistic <- 0
  for (k in 1:n_dims) {
    L <- t(linear_coef[k, ] %*% theta_occasion)
    var_L <- array(0, c(n_dims, n_dims))
    for (t in 1:n_occasions)
      var_L <- var_L + linear_coef[k, t] ^ 2 * solve(obs_fi[, , t])
    test_statistic <- test_statistic + t(L) %*% solve(var_L) %*% L
  }
  
  df <- n_dims * (n_occasions - 1)
  p_value <- pchisq(test_statistic, df, lower.tail = FALSE)
  decision <- ifelse(p_value < alpha, 1, 0)
  return(list(test_statistic = test_statistic, p_value = p_value, decision = decision))
}


ST_omnibus <- function(item_params, resp, theta_pooled, n_dims, n_occasions, D, alpha = 0.05) {
  # compute observed information matrices
  obs_fi <- array(0, c(n_dims, n_dims, n_occasions))
  for (t in 1:n_occasions)
    obs_fi[, , t] <- -log_lik(item_params[, , t], resp[, t], n_dims, D, theta_pooled, order = 2)
  
  OI <- array(0, c(n_dims * n_occasions, n_dims * n_occasions))
  OS <- array(0, c(n_dims * n_occasions))
  
  for (t in 1:n_occasions) {
    index <- (1 + (t - 1) * n_dims):(n_dims * t)
    OI[index, index] <- obs_fi[, , t]
    OS[index] <- log_lik(item_params[, , t], resp[, t], n_dims, D, theta_pooled, order = 1)
  }
  
  test_statistic <- t(OS) %*% solve(OI) %*% OS
  df <- n_dims * (n_occasions - 1)
  p_value <- pchisq(test_statistic, df, lower.tail = FALSE)
  decision <- ifelse(p_value < alpha, 1, 0)
  return(list(test_statistic = test_statistic, p_value = p_value, decision = decision))
}
